--------------------------------------------------------------------------------
--
-- Name:         EuDoc - Euphoria documentation program
-- Version:      0.4
--
-- Author:       D. Newhall
--
-- License:      Mozilla
--
-- Created:      Feb-7-2005
-- Updated:      Feb-14-2005
-- Updated:      Jul-13-2005
-- Updated:      Jul-29-2005 Separated EuDocs.ex to EuDoc.ex and EuLib.e
-- Updated:      Aug-5-2005 Version 0.1
--                           First upload to the Archive
-- Updated:      Aug-12-2005 Version 0.2
--                            Added "Date" tag as synonym for "Created"
--                            Fixed 8.3 filename bug
--                            Added "Variable" tag as synonym for "Routine"
-- Updated:      Sept-18-2005 Version 0.3
--                             Rearranged internal structure of EuDocLib.e
--                             Added new routines and variables for better 
--                              control over using EuDocLib in other programs
--                             Added date created to copyright info generated
--                             Added "Comment" as synonym for "Comments"
--                             Added better prompts and documentation
--                             Fixed bug where routines in the alphabetical
--                              routines listing file weren't sorted correctly
--                              (thanks to Andy Drummond for pointing this out)
--                             Finished debugging routine (print_data)
-- Updated:      Oct-8-2005 Version 0.4
--                           Added "Description" and "Notes" tags
--                           Added "Introduction" as a synonym for "Intro"
--                           Added error checking when generating HTML
--                           Added Show_notes and Ignore_routines output flags
--                           Fixed minor bug in trim dealing with newlines
--                           Fixed bug in open_file where wrong length was used
--
--
--
-- Intro:
--  EuDoc is a program for automatically creating documentation
--  for Euphoria libraries. The file created has the same visual
--  appearance as the library routines list in the official Euphoria
--  documentation. It can also be used to document programs but in a
--  more limited scope.    <br> <br>
--
--  For EuDoc to work you first need to document your program using these tags
--  described below. EuDoc collects everything from one of the required
--  starting tags to the first uncommented line it reaches.
--  The tags are case insensitive and need not be in any order except that
--  "Name&#58;" must come before a block of info relating to the file's info,
--  "Routine&#58;" must come before a block of info relating to a routine,
--  "Type&#58;" must come before a block of info relating to a type,
--  "Constant&#58;" must come before a block of info relating to a constant, and
--  "Variable&#58;" must come before a block of info relating to a variable 
--  (this is only supposed to be used for global variables that are exported
--  like db_fatal_id in EDS). However, it is highly recommended that you use
--  the capitalization scheme and order shown for better consistency and
--  clarity. Also, to show emphasis or to do special formatting you just need
--  to put the desired HTML tags in your comments.
--
--  <pre>
--
--  File descriptor tags:
--
--    Name&#58;            Name of library/program. Required first
--    Author&#58;          Author's name
--    Created&#58;         Date file was created
--    Date&#58;            Synonym for "Created&#58;"
--    Updated&#58;         Date file was updated
--    License&#58;         License library's source code is released under
--    Disclaimer&#58;      Disclaimer on use of the library
--    Intro&#58;           Introduction comments about the library
--    Notes&#58;           Notes on implementation details
--    Note&#58;            Synonym for "Notes&#58;"
--
--  Routine\Type\Constant\Variable descriptor tags:
--
--    Routine&#58;         Name of the routine. Required first
--    Type&#58;            Name of the type. Similar to "Routine&#58;"
--    Constant&#58;        Name of the constant. Similar to "Routine&#58;"
--    Variable&#58;        Name of the variable. Similar to "Routine&#58;"
--    Author&#58;          Author of the routine
--    Created&#58;         Date routine was created
--    Date&#58;            Synonym for "Created&#58;"
--    Updated&#58;         Date routine was updated and contributor who did so
--    Platform&#58;        Platform the routine runs on
--                         (General, WIN32, DOS32, Linux, FreeBSD)
--    Syntax&#58;          An illustration of the usage syntax
--    Description&#58;     A description of what the routine does
--    Comments&#58;        Extra comments about the routine
--    Comment&#58;         Synonym for "Comments&#58;"
--    Notes&#58;           Notes on implementation details
--    Note&#58;            Synonym for "Notes&#58;"
--    Example&#58;         A more example to illustrate usage
--    Example Program&#58; The name/location of an example program if such
--                          a program exists
--    See Also&#58;        Lists other related routines in the library
--
--  </pre>
--
--  To try out EuDoc type "ex EuDoc.ex EuDocLib.e" (remember to substitute "ex"
--  for the name of the interpreter on your platform: "exu" for Linux and
--  FreeBSD and "exw" for Windows). EuDocLib.e is the file that contains all
--  the routines for use in EuDoc.ex. It's commented so as to produce the
--  documentation you are reading right now! Besides the information on running
--  EuDoc this file also contains documentation for the routines used by
--  EuDoc as defined in EuDocLib.e as an example.  <br>  <br>
--
--  Please send all your comments, questions, suggestions, and bug reports to
--  <a href="mailto: derek_newhall@yahoo.com">Derek Newhall</a>.
--
--------------------------------------------------------------------------------


-- Include files

include wildcard.e
include get.e
include sort.e
include machine.e




-- Constants

global constant  EUDOC_VERSION = "0.4",  -- Current version of EuDoc

                 -- Logical true and false
                 TRUE = 1,
	         FALSE = 0,


	         LIBRARY = 1,  -- Denotes that library info is being processed
	         ROUTINE = 2,  -- Denotes that routine info is being processed
	 
	         -- Library info constants (some are shared with routines)
	         NAME = 1,
	         VERSION = 2,
	         AUTHOR = 3,
	         CREATED = 4,
	         UPDATED = 5,
	         LICENSE = 6,
	         DISCLAIMER = 7,
	         INTRO = 8,
	         NOTES = 9,

  	         -- Routine info constants
  	         PLATFORM = 5,
	         SYNTAX = 6,
	         DESCRIPTION = 7,
	         COMMENTS = 8,
	         EXAMPLE = 10,
	         EXAMPLE_PROG = 11,
	         SEE_ALSO = 12,

	         -- Initial data for routines
	         --   (added to Routines when new routine is being parsed)
                 ROUTINE_STUB = {0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0}



-- Global variables


--------------------------------------------------------------------------------
-- Variable: Parsing
-- Syntax: Parsing = i
-- Description: 
--   If the program is parsing data than this is set to
--   TRUE (1). Parsing will (in all known cases) set itself to TRUE when parsing
--   occurs however resetting it to FALSE (0) needs to be done explicitly in 
--   programs using this library.
-- Comments: Might be declared as a local variable in future versions.
-- See also: read_line
--------------------------------------------------------------------------------
global integer Parsing,


--------------------------------------------------------------------------------
-- Variable: Error_id
-- Syntax: Error_id = routine_id(st)
-- Description:
--   If you want to redirect the error messages
--   somewhere else then set this variable to the
--   routine id of the routine you want to handle them
--   instead.
-- Comments: Any routine you redirect to can only be
--   a procedure that takes a single sequence
--   (containing the error message) as an argument.
-- See also: error
--------------------------------------------------------------------------------
               Error_id,


--------------------------------------------------------------------------------
-- Variable: Show_notes
-- Syntax: Show_notes = i
-- Description:
--   If this is set to TRUE (1) then generate_output will print out all notes
--   (signified with the "Notes:" tag) in the generated HTML. It is set to FALSE
--   (0) by default.
-- Comments: Content after the "Notes:" tag is meant to be notes for
--   implementors reading the source code to modify it and are not
--   typically meant to be outputted for general documentation.
-- See also: generate_html
--------------------------------------------------------------------------------
               Show_notes,
               
               
--------------------------------------------------------------------------------
-- Variable: Ignore_routines
-- Syntax: Ignore_routines = i
-- Description:
--   If this is set to TRUE (1) then generate_output will not print out any of
--   the routine information in the generated HTML. It is set to FALSE
--   (0) by default.
-- Comments: This is used for documenting programs where there isn't any routine
--   data that the end-user needs to know about to use the program.
-- See also: generate_html
--------------------------------------------------------------------------------
               Ignore_routines



-- Initialize globals

Error_id = 0
Show_notes = FALSE
Ignore_routines = FALSE

-- Parsing and Parse_type are initialized later for every file




-- Local variables (for the entire file)

sequence Library_header,  -- Library header (initialized in empty_data)
         Routines         -- Routine data (initialized in empty_data)

-- These were formerly global integers but weren't used outside of EuDocLib
integer Parse_type,  -- The type of data that read_line is currently acquiring.
        Last         -- Last type of information being stored.
                     -- Should be set to the constant representing the
                     -- current field in Routines[...] or Library_header
                     -- where data is being written to.





--------------------------------------------------------------------------------
-- Routine:  error
-- Syntax:   error(s)
-- Description:
--   Prints out string passed to it and then aborts with an exit status of 1 
--   unless redefined.
-- Comment:
--   error's behavior can be redefined by setting Error_id to be the routine id
--   of another error message handling procedure.
-- See Also: Error_id
--------------------------------------------------------------------------------
global procedure error(sequence msg)

    if Error_id = 0 then
        printf(2, "Error: %s\n", {msg})
        abort(1)
    else
        call_proc(Error_id, msg)
    end if

end procedure  -- error




--------------------------------------------------------------------------------
-- Routine:  trim
-- Syntax:   s1 = trim(s2)
-- Description:  Removes leading and trailing spaces, tabs, and '-'s from the
--  string passed to it.
--------------------------------------------------------------------------------
global function trim(sequence string)

    sequence tmp
    
    tmp = ""
    
    --  if length(string) = 0 then
    --       return FALSE
    --   end if

    for i=1 to length(string) do
	if string[i] != ' ' and string[i] != '\t' and string[i] != '\n'
          and string[i] != '-' then
	    tmp = string[i..length(string)]
	    exit
	-- Else do next iteration
	end if
    end for

    -- If line is all white-space return FALSE
    if length(tmp) = 0 then
	return FALSE
    end if

    for i=length(tmp) to 1 by -1 do
	if tmp[i] != ' ' and tmp[i] != '\t' and tmp[i] != '\n'
           and tmp[i] != ':' then
	    return tmp[1..i]
	end if
    end for

end function  -- trim




--------------------------------------------------------------------------------
-- Routine:  trim_comment
-- Syntax:   s1 = trim(s2)
-- Description:
--   Same as trim instead it just removes "--" from the beginning of the
--   string.
-- Comments: Only used in read_line.
-- See Also: trim, read_line
--------------------------------------------------------------------------------
global function trim_comment(sequence string)

    sequence tmp
    
    tmp = ""

    for i=1 to length(string) do
	if string[i] != '-' then
	    tmp = string[i..length(string)]
	    exit
	-- Else do next iteration
	end if
    end for

    return tmp

end function  -- trim_comment




--------------------------------------------------------------------------------
-- Routine: empty_data
-- Syntax:  empty_data(i)
-- Description:
--  empty_data is used for processing multiple files one after another.
--  Initializes/resets stored data (Routines and Library_header).
--  Also, takes an integer flag, if the flag is TRUE (1) all
--  parsing flags are set to their defaults and parsing ends (since there's no 
--  parsing on initialization). If the flag is FALSE (0) then it only empties
--  (or initializes) the variables holding the parsed documentation information
--  and parsing continues from the last state it was in.
-- See Also: Parsing
--------------------------------------------------------------------------------
global procedure empty_data(integer reset_flag)

    Routines = {}
    Library_header = repeat(0, 9)
    if reset_flag = TRUE then
        Parse_type = LIBRARY
        Parsing = FALSE
    elsif not reset_flag = FALSE then
        error("Bad flag value sent to empty_data.\n" &
          "Should be TRUE (1) or FALSE (0) only.")
    end if

end procedure  -- empty_data




--------------------------------------------------------------------------------
-- Routine: read_line
-- Syntax:  read_line(sequence command, sequence line)
-- Description:
--   Takes the data tag and content from a line of code and stores
--   the content for documentation.
-- See Also: Parsing, generate_html
--------------------------------------------------------------------------------
global procedure read_line(sequence command, sequence line)

    sequence original_command  -- Holds the original command in case command 
                               -- isn't a valid data tag


    original_command = ""


    -- Check for empty command which means put data in last field accessed
    if length(command) = 0 then -- command = ""
	if Parse_type = LIBRARY then
	    Library_header[Last] &= trim_comment(line)
	else -- Parse_type = ROUTINE
	    if Last = SYNTAX then
                Routines[length(Routines)][SYNTAX] &=
                  (trim_comment(line) & "<br>")
	    else
                Routines[length(Routines)][Last] &= trim_comment(line)
	    end if
	end if

	return -- Stop here if comment is empty

    else
        original_command = command
        command = lower(trim(command))
    end if


    if equal(command, "name") then
	if Parse_type = ROUTINE then
	    error(" \"Name:\" tag used after library name has" &
	      " already been defined.")
	else
	    Library_header[NAME] = line
	    Last = NAME
	    Parsing = TRUE
	end if

    elsif equal(command, "version") then
	if Parse_type = ROUTINE then
	    error(" \"Version:\" tag used after library data has" &
	      " already been defined.")
	else
	    Library_header[VERSION] = line
	    Last = VERSION
	    Parsing = VERSION
	end if

    elsif equal(command, "license") then
	if Parse_type = ROUTINE then
	    error("\"License:\" tag used after all library data" &
	      " should have been defined.")
	else
	    Library_header[LICENSE] = line
	    Last = LICENSE
	    Parsing = TRUE
	end if
	
    elsif equal(command, "disclaimer") then
	if Parse_type = ROUTINE then
	    error("\"Disclaimer:\" tag used after all library data" &
	      " should have been defined.")
	else
	    Library_header[DISCLAIMER] = line
	    Last = DISCLAIMER
	    Parsing = TRUE
	end if

    elsif equal(command, "intro") or equal(command, "introduction") then
	if Parse_type = ROUTINE then
	    error("\"Intro:\" tag used after all library data" &
	      " should have been defined.")
	else
	    Library_header[INTRO] = line
	    Last = INTRO
	    Parsing = TRUE
	end if

    elsif equal(command, "routine") or equal(command, "constant")
      or equal(command, "type") or equal(command, "variable") then
	if Parse_type = LIBRARY then
	    Parse_type = ROUTINE
	end if
	Routines = append(Routines, ROUTINE_STUB)
	Routines[length(Routines)][NAME] = line
	Last = ROUTINE
	Parsing = TRUE

    elsif equal(command, "author") then
	if Parse_type = LIBRARY then
	    Library_header[AUTHOR] = line
	else
	    Routines[length(Routines)][AUTHOR] = line
	end if
	Last = AUTHOR
	Parsing = TRUE

    elsif equal(command, "created") or equal(command, "date") then
	if Parse_type = LIBRARY then
	    Library_header[CREATED] = line
	else
	    Routines[length(Routines)][CREATED] = line
	end if
	Last = CREATED
	Parsing = TRUE

    elsif equal(command, "updated") then
	if Parse_type = LIBRARY then
	    Library_header[UPDATED] = line
	else
	    Routines[length(Routines)][AUTHOR] = line
	end if
	Last = UPDATED
	Parsing = TRUE

    elsif equal(command, "platform") then
	if Parse_type = LIBRARY then
	    error("\"Platform:\" tag can only be used to document routines.")
	else
	    Routines[length(Routines)][PLATFORM] = line
	    Last = PLATFORM
	    Parsing = TRUE
	end if

    elsif equal(command, "syntax") then
	if Parse_type = LIBRARY then
	    error("\"Syntax:\" tag can only be used to document routines.")
	else
	    Routines[length(Routines)][SYNTAX] = (line & "<br>")
	    Last = SYNTAX
	    Parsing = TRUE
	end if


    elsif equal(command, "description") then
	    if Parse_type = LIBRARY then
	    error("\"Description:\" tag can only be used to document routines.")
	else
	    Routines[length(Routines)][DESCRIPTION] = line
	    Last = DESCRIPTION
	    Parsing = TRUE
	end if

    elsif equal(command, "comments") or equal(command, "comment") then
	if Parse_type = LIBRARY then
	    error("\"Comments:\" tag can only be used to document routines.")
	else
	    Routines[length(Routines)][COMMENTS] = line
	    Last = COMMENTS
	    Parsing = TRUE
	end if

    elsif equal(command, "note") or equal(command, "notes") then
	if Parse_type = LIBRARY then
	    Library_header[NOTES] = line
	else
	    Routines[length(Routines)][NOTES] = line
	end if
	Last = NOTES
	Parsing = TRUE

    elsif equal(command, "example") then
	if Parse_type = LIBRARY then
	    error("\"Example:\" tag can only be used to document routines.")
	else
	    Routines[length(Routines)][EXAMPLE] = line
	    Last = EXAMPLE
	    Parsing = TRUE
	end if

    elsif equal(command, "example program") then
	if Parse_type = LIBRARY then
	    error("\"Example Program:\" tag can only be used to document " &
            " routines.")
	else
	    Routines[length(Routines)][EXAMPLE_PROG] = line
	    Last = EXAMPLE_PROG
	    Parsing = TRUE
	end if

    elsif equal(command, "see also") then
	if Parse_type = LIBRARY then
	    error("\"See Also:\" tag can only be used to document routines.")
	else
	    Routines[length(Routines)][SEE_ALSO] = line
	    Last = SEE_ALSO
	    Parsing = TRUE
	end if

    else -- Not a valid data tag so treat as a normal comment
        if Parsing = TRUE then
            if Parse_type = LIBRARY then
                Library_header[Last] &= original_command & trim_comment(line)
            else -- Parse_type = ROUTINE
                Routines[length(Routines)][Last] &= original_command &
                  trim_comment(line)
            end if
        end if

    end if


end procedure  -- read_line




--------------------------------------------------------------------------------
-- Routine: comment
-- Syntax:  i = comment(s)
-- Description: Returns index of the end of the comment characters ("--") if
--   line is a comment, 0 otherwise.
--------------------------------------------------------------------------------
global function comment(sequence line)

    -- Check to see if the line starts with a comment
    for j=1 to length(line) do

        if line[j] = '-' then
            if line[j+1] = '-'  then
                return j+2
            else
	        return 0
            end if
        elsif line[j] != ' ' and line[j] != '\t' then
            return 0
        end if

    end for
    
    return 0

end function  -- comment




--------------------------------------------------------------------------------
-- Routine: see_also_links
-- Syntax:  s = see_also_links(sequence filename, sequence routine_names)
-- Description:
--   Takes the name of a file and a string of routine names and returns a 
--   string of HTML that has links to every routine in the file with that the
--   name passed. Needed for parsing "See Also" info.
-- See Also:  generate_html
--------------------------------------------------------------------------------
function see_also_links(sequence file, sequence string)

    sequence word, -- Holds each name while parsing
             text  -- The HTML text returned
    
    word = ""
    text = ""

    -- Parse string by commas ignoring spaces and "()"s
    for i=1 to length(string) do
        if string[i] = ',' then
            if length(word) > 0 then  -- word is not empty
                text &= sprintf("<a href=\"%s#%s\">%s</a>,\n",
                  {file, trim(word), trim(word)})
                word = ""
            end if
        elsif string[i] != ' ' and string[i] != '(' and string[i] != ')' then
           word &= string[i]
        end if
    end for

    -- Last word not separated by a ','
    if length(word) > 0 then
        text &= sprintf("<a href=\"%s#%s\">%s</a>\n",
          {file, trim(word), trim(word)})
    end if
    
    return text

end function  -- see_also_links




--------------------------------------------------------------------------------
-- Routine: open_file
-- Syntax:  x = open_file(st)
-- Description: 
--   Takes a filename, creates the alphabetical routine listing file, and
--   returns two file handles: one for the main HTML documentation and one for
--   the alphabetical routine listing, which should then be sent to
--   generate_html.    <br><br>
--
--   If, instead, open_file returns an atom that means there was a problem with
--   the filename supplied.    <br><br>
--
--   On an error the return values are:     <br><br>
--   0 signifies the filename is too long (11 car max)                    <br>
--   1 signifies that just the file extension is too long (3 char max)    <br>
--   2 signifies that the main file can not be created                    <br>
--   3 signifies that the alphabetical routine listing can't be created   <br>
--
-- Comments: open_file was originally part of generate_html.
--
-- See Also: generate_html
--------------------------------------------------------------------------------
global function open_file(sequence output_file)

    sequence tmp,  -- String for random temporary data while getting file names
             list_file     -- Name of the HTML file that holds the alphabetical
                           --  list of routines

    integer index,  -- Temporary indeces used for generating the name of the
            index2,  --  second file
            fn1,    -- File handle for first file (output_file)
            fn2    -- File handle for second file (list_file)

    -- Get last directory separator before filename
    tmp = reverse(output_file)

    if platform() = LINUX then
        index2 = match("/", tmp)
    elsif platform() = DOS32 then
        index2 = match("\\", tmp)
    elsif platform() = WIN32 then
        -- Only support '\' for now. Technically '/' is valid in Windows too
        index2 = match("\\", tmp)
    else
        -- Should NEVER reach here!
        error("platform() returns an incorrect/unsupported value.")
    end if


    if index2 > 0 then

        -- Set index2 to work on the original string
        index2 = length(output_file) - index2

        if length(output_file[index2..$]) > 12 then
            return 0
        else

            -- Get last '.'
            index = match(".", tmp)

            -- Need to subtract 1 to get rid of index for '.'
            if (index - 1) > 3 then
                return 1
            end if

        end if

    else -- No directory specified

        if length(output_file) > 12 then
            return 0
        else

            -- Get last '.'
            index = match(".", tmp)

            -- Need to subtract 1 to get rid of index for '.'
            if (index - 1) > 3 then
                return 1
            end if

        end if

    end if

    -- Open file to output the alphabetical list to
    -- (it's the first 4 characters of the file name with "_a_z.html" appended)

    if index2 > 0 then
        tmp = output_file[1..(index2 + 1)]
        list_file = output_file[(index2 + 2)..$]
    else
        tmp = ""
        list_file = output_file
    end if

    if length(list_file) <= 4 then
        list_file = list_file & "_a_z.htm"
    else
        list_file = list_file[1..4] & "_a_z.htm"
    end if


    -- Open files for output

    fn1 = open(output_file, "w")
    if fn1 = -1 then
        return 2
    end if

    fn2 = open(tmp & list_file, "w")
    if fn2 = -1 then
        return 3
    end if

    return {fn1, fn2, list_file}

end function  -- open_file




--------------------------------------------------------------------------------
-- Routine: generate_html
-- Syntax:  generate_html(s)
-- Description:
--   Takes a sequence containing two file numbers corresponding to the files
--   that you want the HTML documentation to be outputted to and then uses the
--   data collected in Library_header and Routines creates the
--   HTML documentation from it that looks like the documentation created by
--   RDS.
--
-- Comments:  Very ugly coding when it comes to outputting the HTML usually.
--
-- See Also: open_file
--------------------------------------------------------------------------------
global procedure generate_html(sequence files)

    sequence list_file,     -- Name of the HTML file that holds the alphabetical
                            --  list of routines
             sorted_names,  -- List of routine names sorted alphabetically
             sorted_indeces -- List of the indeces of the data for each sorted
                            --  routine name


    integer fn1,            -- File handle for first file
            fn2,            -- File handle for second file
            index

    object tmp              -- Temporary object, holds data for testing


    -- Get file data from files

    fn1 = files[1]
    fn2 = files[2]
    list_file = files[3]


    -- Start first file
    puts(fn1, "<html>\n<head>\n")

    -- Put name of library as title
    if sequence(Library_header[NAME]) then
       printf(fn1, "<title>%s</title>\n", {Library_header[NAME]})
    end if

    -- Other header stuff might go here (style, etc.)

    -- End HTML head and start HTML body
    puts(fn1, "</head>\n" &
      "<body bgcolor=\"#FFFFFF\" link=\"#0033BB\" vlink=\"#006699\">\n\n")

    -- Print title of library
    if sequence(Library_header[NAME]) then
        printf(fn1, "<center><font color=\"#006633\" size=\"+2\"><b>\n" &
          "<br>\n%s\n", {Library_header[NAME]})
        if sequence(Library_header[VERSION]) then
            printf(fn1, "<br> <br>\nVersion %s\n", {Library_header[VERSION]})
        end if
        puts(fn1,  "</b></font></center>\n")
    end if

    puts(fn1, "<br>\n<br>\n<br>\n")

    -- Print copyright info (date started and author)
    if sequence(Library_header[CREATED])
      or sequence(Library_header[AUTHOR]) then
        puts(fn1, "<center><font color=\"#FF0099\" face=\"Arial, Helvetica\"" &
            " size=\"+1\">\n(c)")
        if sequence(Library_header[CREATED]) then
            puts(fn1, ' ' & Library_header[CREATED])
        end if
        if sequence(Library_header[AUTHOR]) then
            puts(fn1, ' ' & Library_header[AUTHOR])
        end if
        puts(fn1, "\n</font></center><br>\n")
    end if
    
    puts(fn1,"<br>\n<br>\n")

    -- License info will probably go here eventually

    -- Output disclaimer
    if sequence(Library_header[DISCLAIMER]) then
        printf(fn1, "<center><font color=\"#008080\" size=\"+2\"><b>\n" &
          "Disclaimer: </font> <br> %s </b></center>", 
          {Library_header[DISCLAIMER]})
    end if

    -- Print out introductory test
    puts(fn1, "<center><font color=\"#FF0099\" size=\"+2\">\n<br>\n" &
      "   1. Introduction \n </font></center>\n<br><br>\n")
    if sequence(Library_header[INTRO]) then
        printf(fn1, "%s \n <br>\n", {Library_header[INTRO]})
    end if
      
    -- Only output notes if the user explicitly sets Show_notes to 1
    if Show_notes = TRUE and sequence(Library_header[NOTES]) then
        printf(fn1, "<br><br><center><font color=\"#008080\" size=\"+2\"><b>" &
       "\nNotes: </font> </b></center> <br> %s", {Library_header[NOTES]})
    end if


    -- If the user wants to ignore routines (Ignore_routines is set to TRUE)
    -- skip next section and just output end the HTML file, close it and the
    -- second file, and return (should delete second file first but currently 
    -- no way of easily doing that)
    if Ignore_routines = TRUE then
        puts(fn1, "<br>\n</body>\n</html>")
        close(fn1)
        close(fn2)
        return
    end if


    -- Add horizontal rule to separate introduction from routine listing
    puts(fn1, "<br>\n<hr>\n<br>\n")

    -- Routine names for contents

    puts(fn1, "<center><font color=\"#FF0099\" size=\"+2\">\n" &
      "   2. Routines\n</font></center>\n")


    -- Sort routine names alphabetically and print

    sorted_names = {}
    sorted_indeces = repeat("", length(Routines))

    for i=1 to length(Routines) do
        sorted_names = append(sorted_names, trim(Routines[i][NAME]))
    end for

    sorted_names = sort(sorted_names)


    for i=1 to length(sorted_names) do
        puts(fn1, "&nbsp;&nbsp;")
        printf(fn1, "<a href=\"%s#%s\"><b>%s</b></a><br>\n",
          {list_file, sorted_names[i], sorted_names[i]} )
    end for


    -- End of first page

    puts(fn1, "<p>&nbsp;<p>")
    puts(fn1, "<center><font color=\"#006633\"" &
      "face=\"Arial, Helvetica\" size=\"+1\">\n")
    printf(fn1, "... continue \n<a href=\"%s\">", {list_file})
    puts(fn1, "  3. Alphabetical Listing of All Routines, From A to Z\n</a>\n" &
      "</font></center>\n")
      
    puts(fn1, "\n</body>\n</html>")

    close(fn1)



    -- Output Routine contents on second page


    -- Get Routines indeces for each routine name

    sorted_indeces = repeat("", length(sorted_names))

    for i=1 to length(sorted_names) do
        for j=1 to length(Routines) do
            if equal(sorted_names[i], trim(Routines[j][NAME])) then
                sorted_indeces[i] = j
            end if
        end for
    end for

    -- Print out start of second file

    puts(fn2, "<html>\n<head>\n<title>\n")
    if sequence(Library_header[NAME]) then
        puts(fn2, Library_header[NAME] & '\n')
    end if
    puts(fn2, "</title>\n</head>\n<body>\n")
    puts(fn2, "<center>\n <font color=\"#FF0099\" size=\"+2\">\n <br>\n")
    puts(fn2, "   3. Alphabetical Listing of all Routines\n </font>\n</center>")


    -- Print out Routines data using indeces

    for i=1 to length(sorted_indeces) do

         index = sorted_indeces[i]

         -- Print routine/type/constant name
         tmp = trim(Routines[index][NAME])
         printf(fn2, "<a name=\"%s\"></a><font color=\"#006633\" size=\"+2\">" &
           "<br><center><b>%s</b></center></font><p>\n\n",
           {tmp, tmp})
         -- Start table
         puts(fn2, "<table border=\"0\" cellspacing=\"2\"" & 
           "cellpadding=\"2\">\n\n")

         -- Print syntax
         tmp = Routines[index][SYNTAX]
         if sequence(tmp) then
             if length(tmp) > 0 then
                 printf(fn2, "<tr><td valign=\"top\"><b>Syntax:</b> </td>\n" &
                    "<td><font face = \"Arial, Helvetica\">" &
                    "  %s\n </font>\n </td></tr>\n\n", {tmp})
             end if
         end if

         -- Print description
         tmp = Routines[index][DESCRIPTION]
         if sequence(tmp) then
             if length(tmp) > 0 then
                 printf(fn2, "<tr><td valign=\"top\"><b>Description:</b></td>" &
                    "\n<td> %s\n </td></tr>\n\n", {tmp})
             end if
         end if

         -- Print comments
         tmp = Routines[index][COMMENTS]
         if sequence(tmp) then
             if length(tmp) > 0 then
                 printf(fn2, "<tr><td valign=\"top\"><b>Comments:</b></td>\n" &
                    "<td> %s\n </td></tr>\n\n", {tmp})
             end if
         end if
         
         -- Print notes if Show_notes is TRUE
         if Show_notes = TRUE then
             tmp = Routines[index][NOTES]
             if sequence(tmp) then
                 if length(tmp) > 0 then
                     printf(fn2, "<tr><td valign=\"top\"><b>Notes:</b></td>\n" &
                        "<td> %s\n </td></tr>\n\n", {tmp})
                 end if
             end if
         end if

         -- Print example
         tmp = Routines[index][EXAMPLE]
         if sequence(tmp) then
             if length(tmp) > 0 then
                 printf(fn2, "<tr><td valign=\"top\"><b>Example:</b></td>\n" &
                    "<td><font face = \"Arial, Helvetica\">" &
                    "  <pre>%s</pre>\n </font>\n </td></tr>\n\n", {tmp})
             end if
         end if

         -- Print example program
         tmp = Routines[index][EXAMPLE_PROG]
         if sequence(tmp) then
             if length(tmp) > 0 then
                 printf(fn2, "<tr><td valign=\"top\">\n<b>\n  Example<br>\n" &
                   "  Program:\n</b>\n</td>\n" &
                    "<td>\n <font color=\"#5500FF\">\n" &
                    "  %s\n </td></tr>\n\n", {tmp})
             end if
         end if
         
         -- Print see also info
         tmp = Routines[index][SEE_ALSO]
         if sequence(tmp) then
             if length(tmp) > 0 then
                 -- Change string to string with links
                 tmp = see_also_links(list_file, tmp)
                 printf(fn2, "<tr><td valign=\"top\"><b>See Also:</b></td>\n" &
                    "<td>\n  %s\n </td></tr>\n\n", {tmp})
             end if
         end if
         
         -- End table
         puts(fn2, " </table>\n <br> <br>\n")

    end for


    -- End of second file

    puts(fn2, "</body></html>")

    close(fn2)


end procedure  -- generate_html




--------------------------------------------------------------------------------
-- Routine: print_data
-- Syntax:  print_data(x)
-- Description: 
--   This routine outputs the Euphoria data for Library_header and Routines. If
--   x is a sequence then the data is output to a file with that name, if x is
--   an integer representing an open file then the data is outputted to that
--   file.
--
-- Comments: This routine is only used while debugging.
--------------------------------------------------------------------------------
global procedure print_data(object file)

    integer fn      -- File handle for log file
    sequence data   -- Holds routine data
    
    if sequence(file) then
	fn = open(file, "w")
	if fn = -1 then
	    error(sprintf("File \"%s\" already exists or can not be created" &
	      " for output.", {file}) )
	end if
    else -- file is a file handle
        fn = file
    end if

    -- Print Library_header

    puts(fn, "Library_header:\n\n")
    --print(fn, Library_header)
    puts(fn, "Name: \n" & Library_header[NAME] & "\n\n")
    puts(fn, "Author: \n" & Library_header[AUTHOR] & "\n\n")
    puts(fn, "Date: \n" & Library_header[CREATED] & "\n\n")
    puts(fn, "Updated: \n" & Library_header[UPDATED] & "\n\n")
    puts(fn, "License: \n" & Library_header[LICENSE] & "\n\n")
    --puts(1, "Requires: \n" & Library_header[REQUIRES] & "\n\n") -- TAG REMOVED
    puts(fn, "Disclaimer: \n" & Library_header[DISCLAIMER] & "\n\n")
    puts(fn, "Intro: \n" & Library_header[INTRO] & "\n\n")
    puts(fn, "Notes: \n" & Library_header[NOTES] & "\n\n")

    puts(fn, "\n\n\n\n")
    
    -- Print Routines

    puts(fn, "Routines|Types|Constants|Variables:\n\n")
    --print(fn, Routines)
    for i=1 to length(Routines) do
        data = Routines[i]
        puts(fn, "Name: \n" & data[NAME] & "\n\n")
        puts(fn, "Author: \n" & data[AUTHOR] & "\n\n")
        puts(fn, "Created: \n" & data[CREATED] & "\n\n")
        puts(fn, "Updated: \n" & data[UPDATED] & "\n\n")
        puts(fn, "Platform: \n" & data[PLATFORM] & "\n\n")
        puts(fn, "Syntax: \n" & data[SYNTAX] & "\n\n")
        puts(fn, "Description: \n" & data[DESCRIPTION] & "\n\n")
        puts(fn, "Comments: \n" & data[COMMENTS] & "\n\n")
        puts(fn, "Notes: \n" & data[NOTES] & "\n\n")
        puts(fn, "Example: \n" & data[EXAMPLE] & "\n\n")
        puts(fn, "Example Program: \n" & data[EXAMPLE_PROG] & "\n\n")
        puts(fn, "See Also: \n" & data[SEE_ALSO] & "\n\n")
        puts(fn, "\n\n")
    end for

end procedure  -- print_data

